'''
Auto Encoder with Random Weights

H_Encoder =>   
latent z (how to decide dim and target values) =>  
H_Decoder

The encoder could be a RP (random projection) DR layer. 
'''

class AEWW(object):
    pass # Research in future